SystemOrganization addCategory: #'VMs-Seminar'!
SystemOrganization addCategory: #'VMs-Seminar-Tests'!

!MethodNode methodsFor: '*VMs-Seminar' stamp: 'ag 1/20/2014 11:00'!
finalizePLCCompilationOf: method withLiterals: literals
	
	PLCConfiguration plcCompilationEnabled ifFalse: [ ^ self ].
	literals do: [ :lit |
		(lit isKindOf: PLCStruct) ifTrue: [ lit compiledMethod: method ] ].
	"Let the VM know this is a PLC method."
	PLCConfiguration automaticallySetFlag ifTrue: [ method setFlag ].! !

!MethodNode methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 14:35'!
generate: trailer using: aCompiledMethodClass
	"The receiver is the root of a parse tree. Answer an instance of aCompiledMethodClass.
	 The argument, trailer, is arbitrary but is typically either the reference to the source code
	 that is stored with every CompiledMethod, or an encoding of the method's temporary names."

	| primErrNode blkSize nLits literals stack method |
	self generate: trailer
		using: aCompiledMethodClass
		ifQuick:
			[:m |
			  m	literalAt: 2 put: encoder associationForClass;
				properties: properties.
			^m].
	primErrNode := self primitiveErrorVariableName ifNotNil:
						[encoder fixTemp: self primitiveErrorVariableName].
	encoder supportsClosureOpcodes ifTrue:
		[self ensureClosureAnalysisDone.
		 encoder rootNode: self. "this is for BlockNode>>sizeCodeForClosureValue:"].
	blkSize := (block sizeCodeForEvaluatedValue: encoder)
				+ (primErrNode
					ifNil: [0]
					ifNotNil: [primErrNode sizeCodeForStore: encoder "The VM relies on storeIntoTemp: (129)"]).
	method := aCompiledMethodClass
				newBytes: blkSize
				trailerBytes: trailer 
				nArgs: arguments size
				nTemps: (encoder supportsClosureOpcodes
							ifTrue: [| locals |
									locals := arguments,
											  temporaries,
											  (primErrNode
												ifNil: [#()]
												ifNotNil: [{primErrNode}]).
									encoder
										noteBlockExtent: block blockExtent
										hasLocals: locals.
									locals size]
							ifFalse: [encoder maxTemp])
				nStack: 0
				nLits: (nLits := (literals := encoder allLiterals) size)
				primitive: primitive.
	nLits > 255 ifTrue:
		[ ^self error: 'Too many literals referenced.' ].
	1 to: nLits do: [:lit | method literalAt: lit put: (literals at: lit)].
	encoder streamToMethod: method.
	stack := ParseStack new init.
	primErrNode ifNotNil: [primErrNode emitCodeForStore: stack encoder: encoder].
	stack position: method numTemps.
	block emitCodeForEvaluatedValue: stack encoder: encoder. 
	stack position ~= (method numTemps + 1) ifTrue:
		[^self error: 'Compiler stack discrepancy'].
	encoder methodStreamPosition ~= (method size - trailer size) ifTrue:
		[^self error: 'Compiler code size discrepancy'].
	method needsFrameSize: stack size - method numTemps.
	method properties: properties.
	self finalizePLCCompilationOf: method withLiterals: literals.
	^ method! !

!Encoder methodsFor: '*VMs-Seminar' stamp: 'ag 1/20/2014 10:32'!
encodeSelector: selector
	"Do not reuse SelectorNode instances.
	Required to assign a separate PLC literal for each call site.
	See SelectorNode >> #sizeCode:args:super:, usage of plcIndex."
	
	^(self
		name: selector
		key: selector
		class: SelectorNode
		type: SendType
		set: selectorSet)
			copyWithoutPlcLiteral! !

!CCodeGenerator methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 12:49'!
addMethod: aTMethod
	"Add the given method to the code base."

	(methods includesKey:  aTMethod selector) ifTrue: [
		Transcript cr;
			show: 'Not adding method ', aTMethod selector, ' from class ';
			show: aTMethod definingClass name, '. Already added from ', (methods at: aTMethod selector) definingClass name.
		^ self
		"self error: 'Method name conflict: ', aTMethod selector."
	].
	methods at: aTMethod selector put: aTMethod.! !

!CCodeGenerator methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:51'!
checkClassForNameConflicts: aClass
	"Verify that the given class does not have constant, variable, or method names that conflict with
	 those of previously added classes. Raise an error if a conflict is found, otherwise just return."

	"check for constant name collisions in class pools"
	aClass classPool associationsDo:
		[:assoc |
		(constants includesKey: assoc key asString) ifTrue:
			[self error: 'Constant ', assoc key, ' was defined in a previously added class']].

	"and in shared pools"
	(aClass sharedPools reject: [:pool| pools includes: pool]) do:
		[:pool |
		pool bindingsDo:
			[:assoc |
			(constants includesKey: assoc key asString) ifTrue:
				[self error: 'Constant ', assoc key, ' was defined in a previously added class']]].

	"check for instance variable name collisions"
	(aClass inheritsFrom: VMStructType) ifFalse:
		[aClass instVarNames do:
			[:varName |
			(variables includes: varName) ifTrue:
				[self error: 'Instance variable ', varName, ' was defined in a previously added class']]].

	"check for method name collisions"
	aClass selectors do:
		[:sel |
		((methods includesKey: sel) and:
			[ | meth |
			meth := aClass compiledMethodAt: sel.
			meth isAbstract not and: [(meth pragmaAt: #doNotGenerate) isNil]]) ifTrue:
				["self error: 'Method ', sel, ' was defined in a previously added class.'"]]! !

!EncoderForV3 methodsFor: '*VMs-Seminar' stamp: 'ag 1/7/2014 16:32'!
genBranchPopFalse: distance
	"See BlueBook page 596"
	distance < 0 ifTrue:
		[^self outOfRangeError: 'distance' index: distance range: 0 to: 1023].
	(distance > 0 and: [distance < 9]) ifTrue:
		["152-159 	10011iii 	Pop and Jump 0n False iii +1 (i.e., 1 through 8)"
		 stream nextPut: 152 + distance - 1.
		 ^self].
	distance < 1024 ifTrue:
		["172-175 	101011ii jjjjjjjj 	Pop and Jump On False ii *256+jjjjjjjj"
		 stream
			nextPut: 172 + (distance bitShift: -8);
			nextPut: distance + 1024 \\ 256.
		 ^self].
	PLCConfiguration veryLongJumpCompilationEnabled ifTrue: [ ^ self genVeryLongJump: distance precededBy: 135 ].
	^self outOfRangeError: 'distance' index: distance range: 0 to: 1023! !

!EncoderForV3 methodsFor: '*VMs-Seminar' stamp: 'ag 1/7/2014 16:32'!
genBranchPopTrue: distance
	"See BlueBook page 596"
	distance < 0 ifTrue:
		[^self outOfRangeError: 'distance' index: distance range: 0 to: 1023].
	distance < 1024 ifTrue:
		["168-171 	101010ii jjjjjjjj 	Pop and Jump On True ii *256+jjjjjjjj"
		 stream
			nextPut: 168 + (distance bitShift: -8);
			nextPut: distance + 1024 \\ 256.
		 ^self].
	PLCConfiguration veryLongJumpCompilationEnabled ifTrue: [ ^ self genVeryLongJump: distance precededBy: 135 ].
	^self outOfRangeError: 'distance' index: distance range: 0 to: 1023! !

!EncoderForV3 methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 15:30'!
genJumpLong: distance
	"See BlueBook page 596"
	
	(distance >= -1024 and: [distance < 1024]) ifTrue:
		["160-167 	10100iii jjjjjjjj 	Jump(iii - 4) *256+jjjjjjjj"
		 stream
			nextPut: 160 + (distance + 1024 bitShift: -8);
			nextPut: distance + 1024 \\ 256.
		 ^self].
	PLCConfiguration veryLongJumpCompilationEnabled ifTrue: [ ^ self genVeryLongJump: distance ].
	^self outOfRangeError: 'distance' index: distance range: -1024 to: 1023! !

!EncoderForV3 methodsFor: '*VMs-Seminar' stamp: 'ag 1/19/2014 16:11'!
genPLCStub: plcLiteral
	"First byte: simple jump bytecode to jump over the second byte.
	Second byte: the literal-index that will store a PLCStruct for this callsite."
	
	stream
		nextPut: 144; "jump +1"
		nextPut: plcLiteral.! !

!EncoderForV3 methodsFor: '*VMs-Seminar' stamp: 'ag 1/7/2014 15:20'!
genVeryLongJump: distance
	
	self genVeryLongJump: distance precededBy: 0.! !

!EncoderForV3 methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 13:19'!
genVeryLongJump: distance precededBy: extraBytecode
	"This was added for the VMs-Seminar."
	"We are using the unused bytecode 126 + 2 extra bytes encoding the jump destination."
	
	PLCConfiguration veryLongJumpCompilationEnabled ifFalse: [ ^ self ].
	(distance >= -32768 and: [distance <= 33023]) ifFalse:
		[ ^ self outOfRangeError: 'distance' index: distance range: -32768 to: 33023 ].
	extraBytecode = 0 ifFalse: [
		stream nextPut: extraBytecode ].
	stream
		nextPut: 126;
		nextPut: (distance bitShift: -8) + 128;
		nextPut: distance \\ 256.! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 00:24'!
canHandlePlcLiterals
	"If this method does not contain PLCs yet, it must have enough literals
	available to hold one extra object per callsite."
	
	^ self hasPLCLiteral or: [ 255 > (self numLiterals + self numberOfMessages) ]! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/21/2014 10:06'!
flushCache
	
	PLCConfiguration automaticallyFlushPLCs
		ifTrue: [ self flushPLCCache ].
	self flushOriginalMethodCache.! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/28/2014 12:03'!
flushContainedPLCs
	
	self literals do: [ :lit |
		lit class == PLCStruct
			ifTrue: [ lit flushAll ] ].! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/17/2014 15:32'!
flushOriginalMethodCache
	"Tell the interpreter to remove all references to this method from its method lookup cache, if it has one.  This primitive must be called whenever a method is defined or removed.
	NOTE:  Only one of two selective flush methods needs to be used.
	Squeak 2.2 and earlier uses 119 (See Symbol flushCache).
	Squeak 2.3 and later uses 116 (See CompiledMethod flushCache)."

	<primitive: 116>
! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/20/2014 13:51'!
flushPLCCache
	
	PLCStruct allInstancesDo: [ :struct |
		struct flushMethod: self ].! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/21/2014 23:29'!
hasAnyMessages
	
	| scanner |
	scanner := InstructionStream on: self.
	scanner
		scanFor: 
			[:x |
			scanner selectorToSendOrSelf == scanner
				ifFalse: [ ^ true ].
			false "keep scanning" ].
	^ false! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/20/2014 14:30'!
hasPLCLiteral
	
	^ self literals anySatisfy: [ :lit | lit class == PLCStruct ]! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 00:28'!
hasVeryLongJump
	
	^ (InstructionStream on: self)
		scanFor: [ :x | x = 126 ]! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 13:29'!
isActive
	"Answer whether the receiver is currently the active method inside it's classes methodDict."
	
	^ self == (self methodClass methodDict at: self selector ifAbsent: [ ^ false ])! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 00:32'!
numberOfMessages
	"This is copied from #messages in CompiledMethod, but instead of a Set,
	we collect the messages into a OrderedCollection to count the actual callsites."
	
	| messages scanner |
	messages := 0.
	scanner := InstructionStream on: self.
	scanner
		scanFor: 
			[:x |
			scanner selectorToSendOrSelf == scanner
				ifFalse: [ messages := messages + 1 ].
			false "keep scanning" ].
	^ messages! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/29/2014 10:58'!
plcAtBytecode: bytecode
	
	| literalIndex plc |
	self flag ifFalse: [
		^ self error: 'This method is not flagged (see #setFlag), PLCs are not in use.' ].
	literalIndex := (self at: bytecode + 2) + 1.
	plc := self literalAt: literalIndex.
	plc class == PLCStruct ifFalse: [
		^ self error: 'Unexpected literal object at index ', literalIndex asString ].
	^ plc! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/19/2014 15:33'!
recompile
	
	^ self methodClass recompile: self selector! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/13/2014 16:07'!
setFlag
	"Set the user-level flag bit"
	
	self objectAt: 1 put: (self header bitOr: (1 << 29))! !

!CompiledMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/29/2014 10:58'!
typesAtBytecode: bytecode
	
	^ (self plcAtBytecode: bytecode)
		receiverClasses reject: #isNil! !

!Behavior methodsFor: '*VMs-Seminar' stamp: 'ag 1/17/2014 15:03'!
recompileAll
	
	self selectors do: [ :sel | self recompile: sel ].! !

!SelectorNode methodsFor: '*VMs-Seminar' stamp: 'ag 1/20/2014 10:31'!
copyWithoutPlcLiteral
	
	| copy |
	copy := self copy.
	copy forgetPlcLiteral.
	^ copy! !

!SelectorNode methodsFor: '*VMs-Seminar' stamp: 'ag 1/16/2014 14:59'!
emitCode: stack args: nArgs encoder: encoder super: supered
	| result |
	stack pop: nArgs.
	result := supered
		ifTrue:
			[encoder genSendSuper: index numArgs: nArgs]
		ifFalse:
			[encoder
				genSend: (code < Send ifTrue: [code negated] ifFalse: [index])
				numArgs: nArgs].
	PLCConfiguration plcCompilationEnabled ifTrue: [ encoder genPLCStub: plcLiteral ].
	^ result! !

!SelectorNode methodsFor: '*VMs-Seminar' stamp: 'ag 1/20/2014 10:31'!
forgetPlcLiteral
	
	plcLiteral := nil.! !

!SelectorNode methodsFor: '*VMs-Seminar' stamp: 'ag 1/20/2014 10:14'!
sizeCode: encoder args: nArgs super: supered
	| size |
	self reserve: encoder.
	size := supered
		ifTrue:
			[code < Send "i.e. its a special selector" ifTrue:
				[code := self code: (index := encoder sharableLitIndex: key) type: 5].
			 encoder sizeSendSuper: index numArgs: nArgs]
		ifFalse:
			[self flag: #yuck. "special selector sends cause this problem"
			 encoder
				sizeSend: (code < Send ifTrue: [code negated] ifFalse: [index])
				numArgs: nArgs].
	PLCConfiguration plcCompilationEnabled
		ifTrue: [
			"At a PLC callsite, we store 2 bytes: one to jump over the second byte
			in case a non-modified VM is used, and a second one storing the literal-index for the plc-pointer.
			Allocate the PLCStruct now, at compile-time. Less work for the VM."
			size := size + 2.
			plcLiteral ifNil: [
				plcLiteral := encoder litIndex: (PLCStruct forSelector: key) ] ].
	^ size! !

ClassicObjectMemory subclass: #PLCObjectMemory
	instanceVariableNames: 'classPlcStruct plcStructClassAvailable'
	classVariableNames: 'ClassPLCStruct'
	poolDictionaries: ''
	category: 'VMs-Seminar'!

!PLCObjectMemory commentStamp: '<historical>' prior: 0!
ClassicObjectMemory represents the original Squeak object memory used by Interpreter. Some methods from ObjectMemory are modified or extended in other object memory implementations such as NewObjectMemory. The original implementations of these methods have been moved to ClassicObjectMemory in order to retain the original Squeak object memory while enabling variations such as NewObjectMemory to be implemented as subclasses of ObjectMemory.!

!PLCObjectMemory class methodsFor: 'translation' stamp: 'ag 1/14/2014 13:55'!
declareCVarsIn: aCCodeGenerator
	
	super declareCVarsIn: aCCodeGenerator.
	self declareCAsOop: { #classPlcStruct. #plcStructClassAvailable } in: aCCodeGenerator.! !

!PLCObjectMemory class methodsFor: 'initialization' stamp: 'ag 1/13/2014 20:34'!
initializeSpecialObjectIndices
	
	super initializeSpecialObjectIndices.
	ClassPLCStruct := 46.! !

!PLCObjectMemory class methodsFor: 'translation' stamp: 'ag 1/13/2014 20:19'!
interpreterClass
	
	^ PLCInterpreter! !

!PLCObjectMemory methodsFor: 'plc' stamp: 'ag 1/13/2014 21:05'!
classPLCStruct
	
	<inline: true>
	^ classPlcStruct! !

!PLCObjectMemory methodsFor: 'plc' stamp: 'ag 1/14/2014 13:26'!
hasPLCStruct
	"This is called very frequently. Use an extra variabl to store this boolean."
	
	<inline: true>
	^ plcStructClassAvailable! !

!PLCObjectMemory methodsFor: 'initialization' stamp: 'ag 1/17/2014 18:37'!
initializeObjectMemory: bytesToShift
	
	super initializeObjectMemory: bytesToShift.
	classPlcStruct := self splObj: ClassPLCStruct.
	plcStructClassAvailable := classPlcStruct ~= nilObj.
	
	"Notify user that PLCStruct is not contained at the expected index in the specialObjectsArray."
	plcStructClassAvailable ifFalse: [
		interpreter
			print: 'PLCStruct not found. Execute ''PLCConfiguration fixSpecialObjectsArray'', save image and restart VM.';
			cr ].! !

!PLCObjectMemory methodsFor: 'plc' stamp: 'ag 1/20/2014 11:20'!
isOldNonIntegerOop: oop
	
	^ self oop: oop isLessThan: youngStart! !

Object subclass: #PLCConfiguration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar'!

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/21/2014 10:04'!
automaticallyFlushPLCs
	"If this returns true, All PLCStruct instances will be iterated on each recompilation of a method.
	They will flush the old method from their cache, if it was present.
	Due to the large number of PLCStructs, this can take a long time and can be disabled in some cases."
	
	^ Smalltalk at: #'VMs-Seminar-automaticallyFlushPLCs' ifAbsent: [ false ]! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/19/2014 17:10'!
automaticallySetFlag
	"If this returns true, every method compiled with PLC-support will also be flagged,
	so that the VM will actually use the PLCs. Should only be enabled when PLC-debugging is off.
	Will cause a flood of PLC-operations and debug outputs.
	Only has any effect, if plcCompilationEnabled is enabled."
	
	^ Smalltalk at: #'VMs-Seminar-automaticallySetFlag' ifAbsent: [ false ]! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/22/2014 00:30'!
disableAll
	
	self disableLongJumpCompilation.
	self disablePlcCompilation.
	self disableAutomaticallySetFlag.
	"Do not disable autmatic PLC flushing. Too dangerous."! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/21/2014 10:15'!
disableAutomaticallyFlushPLCs
	"self disableAutomaticallyFlushPLCs"
	"Disable this is dangerous and can lead to weird behaviour as overwritten CompiledMethod instances might be executed.
	Flush the cashe before."
	
	PLCStruct flushAll.
	self disableAutomaticallySetFlag.
	^ Smalltalk globals removeKey: #'VMs-Seminar-automaticallyFlushPLCs' ifAbsent: [ "nothing" ]! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/21/2014 10:13'!
disableAutomaticallySetFlag
	"self disableAutomaticallySetFlag"
	
	^ Smalltalk globals removeKey: #'VMs-Seminar-automaticallySetFlag' ifAbsent: [ "nothing" ]! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/7/2014 16:56'!
disableLongJumpCompilation
	"self disableLongJumpCompilation"
	
	^ Smalltalk globals removeKey: #'VMs-Seminar-veryLongJumpCompilationEnabled' ifAbsent: [ "nothing" ]! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/7/2014 16:56'!
disablePlcCompilation
	"self disablePlcCompilation"
	
	^ Smalltalk globals removeKey: #'VMs-Seminar-plcCompilationEnabled' ifAbsent: [ "nothing" ]! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/21/2014 23:29'!
enableAll
	
	self enableLongJumpCompilation.
	self enablePlcCompilation.
	self enableAutomaticallySetFlag.
	self enableAutomaticallyFlushPLCs.! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/21/2014 10:06'!
enableAutomaticallyFlushPLCs
	"self enableAutomaticallyFlushPLCs"
	
	^ Smalltalk at: #'VMs-Seminar-automaticallyFlushPLCs' put: true! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/19/2014 17:10'!
enableAutomaticallySetFlag
	"self enableAutomaticallySetFlag"
	
	^ Smalltalk at: #'VMs-Seminar-automaticallySetFlag' put: true! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/7/2014 16:32'!
enableLongJumpCompilation
	"self enableLongJumpCompilation"
	
	^ Smalltalk at: #'VMs-Seminar-veryLongJumpCompilationEnabled' put: true! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/7/2014 16:41'!
enablePlcCompilation
	"self enablePlcCompilation"
	
	^ Smalltalk at: #'VMs-Seminar-plcCompilationEnabled' put: true! !

!PLCConfiguration class methodsFor: 'fixing special objects array' stamp: 'ag 1/14/2014 14:25'!
fixSpecialObjectsArray
	"This should be done before opening an image with a PLCInterpreter, otherwise it will not be able to handle PLCs."
	"self fixSpecialObjectsArray"
	
	Smalltalk specialObjectsArray at: 47 put: PLCStruct.! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/7/2014 16:34'!
plcCompilationEnabled
	"PLC-compilation affects two separate places in the compiler,
	which both have to be enabled at the same time."
	
	^ Smalltalk at: #'VMs-Seminar-plcCompilationEnabled' ifAbsent: [ false ]! !

!PLCConfiguration class methodsFor: 'misc' stamp: 'ag 1/20/2014 10:18'!
printStack
	
	| c |
	c := thisContext.
	Transcript cr; show: ' ===== STACK ===== '.
	[ c isNil ] whileFalse: [
		c := c sender.
		Transcript cr.
		c printOn: Transcript ].! !

!PLCConfiguration class methodsFor: 'misc' stamp: 'ag 1/22/2014 00:33'!
recompileAllMethods
	
	| classesToAvoid |
	classesToAvoid := Behavior withAllSuperclasses, Morph withAllSubclasses.
	(Smalltalk allClasses copyWithoutAll: classesToAvoid)
	do: [ :cl | | selectors |
		Transcript cr; show: 'Recompiling class ', cl name.
		selectors := cl selectors.
		selectors do: [ :sel |
			(cl >> sel) flag
				ifTrue: [ Transcript nextPut: $-; flush ]
				ifFalse: [
					(cl >> sel) canHandlePlcLiterals
						ifTrue: [
							cl recompile: sel.
							Transcript nextPut: $.; flush ]
						ifFalse: [ Transcript nextPut: $x; flush ] ] ] ].! !

!PLCConfiguration class methodsFor: 'misc' stamp: 'ag 1/15/2014 11:38'!
regenerateVM
	"Grab the first (should be the only) VMMaker instance and tell it to regenerate the interp.c file."
	
	| vm |
	Smalltalk garbageCollect.
	vm := VMMaker allSubInstances first.
	(vm coreVMDirectory fileExists: vm interpreterFilename)
		ifTrue: [ vm coreVMDirectory deleteFileNamed: vm interpreterFilename ].
	vm generateInterpreterFile.! !

!PLCConfiguration class methodsFor: 'fixing special objects array' stamp: 'ag 1/14/2014 14:30'!
unfixSpecialObjectsArray
	"self unfixSpecialObjectsArray"
	
	Smalltalk specialObjectsArray at: 47 put: nil.! !

!PLCConfiguration class methodsFor: 'config-flags' stamp: 'ag 1/7/2014 16:32'!
veryLongJumpCompilationEnabled
	
	^ Smalltalk at: #'VMs-Seminar-veryLongJumpCompilationEnabled' ifAbsent: [ false ]! !

Object subclass: #VmBenchmarks
	instanceVariableNames: 'random syntheticalTestObjects syntheticalTestClasses'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar'!

!VmBenchmarks class methodsFor: 'polymorphic' stamp: 'ag 12/6/2013 11:07'!
a
	^ 1! !

!VmBenchmarks class methodsFor: 'polymorphic' stamp: 'ag 12/6/2013 11:07'!
b
	^ 2! !

!VmBenchmarks class methodsFor: 'bench' stamp: 'ag 12/6/2013 12:39'!
bench: t1 forSeconds: t2 
	| t3 t4 t5 |
	t3 := 0.
	t5 := Time millisecondClockValue + (t2 * 1000).
	t4 := Time millisecondClockValue.
	[Time millisecondClockValue > t5]
		whileFalse: [t1 value.
			t3 := t3 + 1].
	t5 := Time millisecondClockValue.
	^ (t3 * 1000 / (t5 - t4)) asFloat! !

!VmBenchmarks class methodsFor: 'bench' stamp: 'ag 12/6/2013 12:42'!
bench: block forSeconds: seconds times: number
	"Will measure how often block can be executed per second. Will run this <number> times.
	The total runtime of this is seconds * number.
	Example:
	VmBenchmarks bench: [ VmBenchmark benchFibonacci ] forSeconds: 5 times: 10
	This will run for 50 seconds."
	
	| times |
	times := Set new.
	number timesRepeat: [ times add: (self bench: block forSeconds: seconds) ].
	^ times average! !

!VmBenchmarks class methodsFor: 'fibonacci' stamp: 'ag 12/6/2013 12:39'!
benchFibonacci
	^ self fibonacci: 1000! !

!VmBenchmarks class methodsFor: 'polymorphic' stamp: 'ag 12/6/2013 11:07'!
benchPolymorphic
	1000 timesRepeat: [ self a + self b + self c + self d + self e + self f ]! !

!VmBenchmarks class methodsFor: 'polymorphic' stamp: 'ag 12/6/2013 11:07'!
c
	^ 3! !

!VmBenchmarks class methodsFor: 'polymorphic' stamp: 'ag 12/6/2013 11:07'!
d
	^ 4! !

!VmBenchmarks class methodsFor: 'polymorphic' stamp: 'ag 12/6/2013 11:07'!
e
	^ 5! !

!VmBenchmarks class methodsFor: 'polymorphic' stamp: 'ag 12/6/2013 11:07'!
f
	^ 6! !

!VmBenchmarks class methodsFor: 'fibonacci' stamp: 'ag 12/6/2013 10:41'!
fibonacci: aNumber
	aNumber < 0 ifTrue: [^0].
	aNumber < 2 ifTrue: [^aNumber].
	^ aNumber + (self fibonacci: aNumber - 1)! !

!VmBenchmarks methodsFor: 'bench synthetical - classes' stamp: 'ag 1/13/2014 14:33'!
allSelectors: aNumber
	
	^ (0 to: aNumber) collect: [ :num | ('test', num asString) asSymbol ]! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:45'!
benchAllSynthetical
	"Measure the minimum runtime of all 4 synthetical benchmarks."
	
	^ #(benchSynthetical3Classes3Methods
		benchSynthetical20Classes3Methods
		benchSynthetical3Classes20Methods
		benchSynthetical20Classes20Methods)
			collect: [ :bench | bench -> (self perform: bench) ]! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:55'!
benchSynthetical20Classes20Methods
	
	^ self
		benchSynthetical: #invoke20Methods
		repeat: 10
		numberOfClasses: 20! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:55'!
benchSynthetical20Classes3Methods
	
	^ self
		benchSynthetical: #invoke3Methods
		repeat: 30
		numberOfClasses: 20! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:55'!
benchSynthetical3Classes20Methods
	
	^ self
		benchSynthetical: #invoke20Methods
		repeat: 50
		numberOfClasses: 3! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:56'!
benchSynthetical3Classes3Methods
	
	^ self
		benchSynthetical: #invoke3Methods
		repeat: 100
		numberOfClasses: 3! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:56'!
benchSynthetical: methodName repeat: repeatCount numberOfClasses: numClasses
	
	self ensureSetup.
	self setupSynthetical: methodName numberOfClasses: numClasses.
	^ self minimumTime: methodName repeat: repeatCount * self repeatFactor! !

!VmBenchmarks methodsFor: 'bench synthetical - classes' stamp: 'ag 1/13/2014 14:35'!
createClass: name
	
	^ Object subclass: name
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: ''
		category: 'VMs-Seminar-Tests'! !

!VmBenchmarks methodsFor: 'bench synthetical - classes' stamp: 'ag 1/13/2014 14:35'!
createClasses: aNumber
	
	^ (1 to: aNumber) collect: [ :num |
		self createClass: #'VmsSeminarGeneratedTestClass', num asString ]! !

!VmBenchmarks methodsFor: 'bench synthetical - classes' stamp: 'ag 1/13/2014 14:37'!
createClasses: classes withMethods: methods
	
	^ (self createClasses: classes)
		collect: [ :cl |
			self createMethods: methods in: cl.
			cl ]! !

!VmBenchmarks methodsFor: 'bench synthetical - classes' stamp: 'vk 1/27/2014 13:49'!
createMethod: selector in: class
	
	^ class compile: (selector asString, ' ^ ', (random nextInt: 1000000))! !

!VmBenchmarks methodsFor: 'bench synthetical - classes' stamp: 'vk 1/27/2014 13:41'!
createMethods: aNumber in: class 
	^ (0 to: aNumber)
		collect: [:num | self createMethod: 'test' , num asString in: class]! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:37'!
ensureSetup
	
	syntheticalTestClasses
		ifNil: [ self setupSyntheticalClasses: 20 numberOfMethods: 20 ].! !

!VmBenchmarks methodsFor: 'initialize' stamp: 'ag 1/31/2014 14:21'!
finalize
	
	super finalize.
	self tearDown.! !

!VmBenchmarks methodsFor: 'initialize' stamp: 'ag 1/31/2014 14:33'!
initialize
	
	super initialize.
	random := Random new.! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 13:48'!
invoke20Methods
	
	1 to: syntheticalTestObjects size do:
		[ :index |
			(syntheticalTestObjects at: index)
			test1;
			test2;
			test3;
			test4;
			test5;
			test6;
			test7;
			test8;
			test9;
			test10;
			test11;
			test12;
			test13;
			test14;
			test15;
			test16;
			test17;
			test18;
			test19;
			test20 ].! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 13:48'!
invoke3Methods
	
	1 to: syntheticalTestObjects size do:
		[ :index |
			(syntheticalTestObjects at: index)
			test1;
			test2;
			test3 ].! !

!VmBenchmarks methodsFor: 'timing' stamp: 'ag 1/31/2014 14:37'!
minimumTime: aSymbol repeat: repeatCount
	"Performing aSymbol should take something like 50-300 ms."
	
	^ self
		minimumTime: aSymbol
		repeat: repeatCount
		testExecutions: 10
		maxExecutions: 50! !

!VmBenchmarks methodsFor: 'timing' stamp: 'ag 1/31/2014 14:52'!
minimumTime: aSymbol repeat: repeatCount testExecutions: testExecutions maxExecutions: maxExecutions
	
	| minTime executions slowerExecutions |
	executions := slowerExecutions := 0.
	minTime := self timeHighPriority: aSymbol repeat: repeatCount.
	Transcript cr; show: '== Started ', aSymbol, ' at ', minTime asString.
	[ slowerExecutions < testExecutions & (executions < maxExecutions) ]
		whileTrue: [ | time |
			time := self timeHighPriority: aSymbol repeat: repeatCount.
			time < minTime
				ifTrue: [
					Transcript cr; show: '- slowerExecutions reset due to ', time asString.
					minTime := time. slowerExecutions := 0 ]
				ifFalse: [ slowerExecutions := slowerExecutions + 1 ].
			executions := executions + 1 ].
	^ minTime! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:56'!
repeatFactor
	
	^ 1000! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:39'!
setupSynthetical: methodName numberOfClasses: numClasses
	
	syntheticalTestObjects := (syntheticalTestClasses copyFrom: 1 to: numClasses) collect: #new.
	(self class >> methodName) flushContainedPLCs.
	"No warmup necessary due to minimumTime:"! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:34'!
setupSyntheticalClasses: numClasses numberOfMethods: numMethods
	
	self tearDown.
	syntheticalTestClasses := self createClasses: numClasses withMethods: numMethods.! !

!VmBenchmarks methodsFor: 'bench synthetical' stamp: 'ag 1/31/2014 14:35'!
tearDown
	
	syntheticalTestClasses ifNotNil: [ syntheticalTestClasses do: #removeFromSystem ].
	syntheticalTestClasses := nil.
	syntheticalTestObjects := nil.! !

!VmBenchmarks methodsFor: 'timing' stamp: 'ag 1/31/2014 14:26'!
time: aSymbol repeat: repeatCount
	"A symbol for perform: is used instead of a block, because
	perform: is faster than activating a block (less overhead measured)."
	
	| time repeatedCount |
	repeatedCount := 0.
	time := Time millisecondClockValue.
	[ repeatedCount < repeatCount ]
		whileTrue: [
			repeatedCount := repeatedCount + 1.
			self perform: aSymbol ].
	^ (Time millisecondClockValue - time) max: 0! !

!VmBenchmarks methodsFor: 'timing' stamp: 'ag 1/31/2014 14:30'!
timeHighPriority: aSymbol repeat: repeatCount
	
	| currentPriority |
	currentPriority := Processor activeProcess priority.
	Processor activeProcess priority: Processor timingPriority - 1.
	[ ^ self time: aSymbol repeat: repeatCount ]
		ensure: [ Processor activeProcess priority: currentPriority ].! !

Object subclass: #VmsSeminarGeneratedTestClass1
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 762075! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 179285! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 970360! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 824189! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 139354! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 112448! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 910311! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 587851! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 995147! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 427702! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 387094! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 872487! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 239782! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 878845! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 658! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 46670! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 378755! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 732908! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 984406! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 907056! !

!VmsSeminarGeneratedTestClass1 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 884809! !

Object subclass: #VmsSeminarGeneratedTestClass10
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 206276! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 879881! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 996473! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 712467! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 427363! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 688878! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 971233! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 511467! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 214095! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 284292! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 93884! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 896268! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 155547! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 562578! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 272037! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 110889! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 696349! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 527194! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 549169! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 871183! !

!VmsSeminarGeneratedTestClass10 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 958648! !

Object subclass: #VmsSeminarGeneratedTestClass11
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 236275! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 69357! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 118476! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 212784! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 254675! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 317045! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 568985! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 923140! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 213011! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 65436! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 776083! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 623231! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 679536! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 634826! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 955084! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 82196! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 462218! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 486435! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 498897! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 949561! !

!VmsSeminarGeneratedTestClass11 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 254961! !

Object subclass: #VmsSeminarGeneratedTestClass12
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 506881! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 139729! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 794027! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 197023! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 355320! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 855630! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 560490! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 148423! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 545304! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 917014! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 239733! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 185447! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 411298! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 807582! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 679115! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 885681! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 628922! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 286785! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 995271! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 508687! !

!VmsSeminarGeneratedTestClass12 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 495734! !

Object subclass: #VmsSeminarGeneratedTestClass13
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 17185! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 817867! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 575469! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 894970! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 756873! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 755555! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 610343! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 20373! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 402482! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 500596! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 505235! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 474132! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 883423! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 721697! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 680290! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 623495! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 63721! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 951964! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 657927! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 769053! !

!VmsSeminarGeneratedTestClass13 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 462164! !

Object subclass: #VmsSeminarGeneratedTestClass14
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 557977! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 902692! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 3385! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 880972! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 492010! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 208180! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 865102! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 759633! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 149865! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 778490! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 78869! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 534717! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 533282! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 980079! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 860891! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 993477! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 364870! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 365518! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 247615! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 649760! !

!VmsSeminarGeneratedTestClass14 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 513953! !

Object subclass: #VmsSeminarGeneratedTestClass15
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 175910! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 512731! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 449295! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 298802! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 961507! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 43446! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 189186! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 639325! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 134872! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 777172! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 917682! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 476122! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 462327! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 171725! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 328363! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 785798! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 906534! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 104859! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 355033! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 38327! !

!VmsSeminarGeneratedTestClass15 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 158057! !

Object subclass: #VmsSeminarGeneratedTestClass16
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 169105! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 139439! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 888265! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 65702! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 245558! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 91746! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 966620! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 965570! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 320903! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 404467! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 871499! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 269325! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 535525! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 541071! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 557867! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 59191! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 807285! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 27248! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 954903! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 42098! !

!VmsSeminarGeneratedTestClass16 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 529178! !

Object subclass: #VmsSeminarGeneratedTestClass17
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 771106! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 978182! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 743464! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 389821! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 713535! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 377674! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 550554! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 158850! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 777242! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 101272! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 67361! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 120206! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 291031! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 287206! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 354627! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 200335! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 14102! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 12030! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 174739! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 832973! !

!VmsSeminarGeneratedTestClass17 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 774841! !

Object subclass: #VmsSeminarGeneratedTestClass18
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 61520! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 952232! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 122231! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 335497! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 684249! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 171018! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 297849! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 939325! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 225870! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 195562! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 796001! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 382605! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 162222! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 436125! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 449187! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 482036! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 568594! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 357584! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 899462! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 249436! !

!VmsSeminarGeneratedTestClass18 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 263112! !

Object subclass: #VmsSeminarGeneratedTestClass19
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 944145! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 233425! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 580073! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 276795! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 91874! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 125406! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 691362! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 714950! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 163606! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 722430! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 876486! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 92784! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 173214! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 420284! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 199070! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 768891! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 741545! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 146479! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 861658! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 880713! !

!VmsSeminarGeneratedTestClass19 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 142952! !

Object subclass: #VmsSeminarGeneratedTestClass2
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 743496! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 921740! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 689879! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 780216! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 85960! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 721970! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 142679! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 2062! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 642234! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 25472! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 106688! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 94140! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 679306! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 195785! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 93397! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 706771! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 684140! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 338341! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 488613! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 107842! !

!VmsSeminarGeneratedTestClass2 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 500428! !

Object subclass: #VmsSeminarGeneratedTestClass20
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 702286! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 305983! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 845847! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 142121! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 621438! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 501668! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 531132! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 727765! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 545515! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 466726! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 255848! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 24172! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 650171! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 244117! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 415324! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 340347! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 210137! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 764842! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 685802! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 265198! !

!VmsSeminarGeneratedTestClass20 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 172657! !

Object subclass: #VmsSeminarGeneratedTestClass3
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 556485! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 842892! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 391369! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 733912! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 851140! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 98888! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 4356! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 208070! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 21812! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 585157! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 717566! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 123469! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 481781! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 140811! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 287383! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 30562! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 653694! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 619671! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 797834! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 192290! !

!VmsSeminarGeneratedTestClass3 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 801357! !

Object subclass: #VmsSeminarGeneratedTestClass4
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 610275! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 890934! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 459692! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 41671! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 349384! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 94490! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 77200! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 494284! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 419653! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 93470! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 940324! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 12619! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 921199! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 71199! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 585339! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 787397! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 768566! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 283600! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 455321! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 577671! !

!VmsSeminarGeneratedTestClass4 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 910363! !

Object subclass: #VmsSeminarGeneratedTestClass5
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 629156! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 215967! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 108673! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 462295! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 790112! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 405957! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 906105! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 900686! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 821048! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 339247! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 711621! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 207257! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 752581! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 367787! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 627063! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 42565! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 388966! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 342764! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 822623! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 813545! !

!VmsSeminarGeneratedTestClass5 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 245262! !

Object subclass: #VmsSeminarGeneratedTestClass6
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 392282! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 71569! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 131395! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 355117! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 451347! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 779555! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 967796! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 745521! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 955153! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 248701! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 906465! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 948497! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 847613! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 383760! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 828612! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 472542! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 6683! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 316586! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 848381! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 736666! !

!VmsSeminarGeneratedTestClass6 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 141140! !

Object subclass: #VmsSeminarGeneratedTestClass7
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 840762! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 683765! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 125863! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 377647! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 108224! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 906139! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 475642! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 113116! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 128636! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 984913! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 419587! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 985452! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 24433! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 477967! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 641104! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 25372! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 423799! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 786631! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 894564! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 935918! !

!VmsSeminarGeneratedTestClass7 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 969366! !

Object subclass: #VmsSeminarGeneratedTestClass8
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 189129! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 678772! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 894669! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 698714! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 277692! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 157424! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 816061! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 535244! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 835350! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 710909! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 244214! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 492494! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 107519! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 343409! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 69163! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 411041! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 354773! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 653909! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 242223! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 36143! !

!VmsSeminarGeneratedTestClass8 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 449747! !

Object subclass: #VmsSeminarGeneratedTestClass9
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar-Tests'!

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test0 ^ 660516! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test1 ^ 290264! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test10 ^ 948138! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test11 ^ 348330! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test12 ^ 370831! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test13 ^ 549869! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test14 ^ 636798! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test15 ^ 661698! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test16 ^ 145488! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test17 ^ 215292! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test18 ^ 398382! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test19 ^ 602461! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test2 ^ 450595! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test20 ^ 554127! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test3 ^ 142160! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test4 ^ 270526! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test5 ^ 719666! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test6 ^ 414134! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test7 ^ 335107! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test8 ^ 129117! !

!VmsSeminarGeneratedTestClass9 methodsFor: 'as yet unclassified' stamp: 'ag 1/31/2014 14:46'!
test9 ^ 65803! !

Array variableSubclass: #PLCStruct
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar'!

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/30/2014 12:42'!
cacheEntryCounterOffset
	"Offset within a cache-entry that contains a counter,
	incremented by the VM every time that cache entry is used."
	
	^ 3! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/22/2014 13:39'!
compiledMethodIndex
	"Index inside the PLCStruct that stores a back-reference to the containing CompiledMethod"
	
	^ 2! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/19/2014 16:08'!
defaultNumberOfEntries
	
	^ 3! !

!PLCStruct class methodsFor: 'instance creation' stamp: 'ag 1/19/2014 16:08'!
defaultSize
	"These numbers correspond to constants in PLCInterpreter."
	
	^ self metaInfoEntries + (self defaultNumberOfEntries * self pointersPerEntry)! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/22/2014 13:40'!
firstEntryIndex
	"Index inside a PLCStruct where the first cache-entry starts."
	
	^ self metaInfoEntries + 1! !

!PLCStruct class methodsFor: 'flushing' stamp: 'ag 1/20/2014 15:22'!
flushAll
	
	PLCStruct allInstances do: #flushAll.! !

!PLCStruct class methodsFor: 'instance creation' stamp: 'ag 1/19/2014 16:14'!
forSelector: selector
	
	^ (self new: self defaultSize)
		selector: selector;
		yourself! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/22/2014 13:40'!
lookupClassOffset
	"Offset within a cache-entry that contains the class for the cache-entry."
	
	^ 0! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/22/2014 13:41'!
messageSelectorIndex
	"Index inside a PLCStruct that contains the selector of the callsite associated with the PLCStruct."
	
	^ 1! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/19/2014 15:08'!
metaInfoEntries
	"Stored at the beginning of a PLCStruct: selector, compiledMethod"
	
	^ 2! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/22/2014 13:41'!
newMethodOffset
	"Offset within a cache-entry that contains the cached CompiledMethod."
	
	^ 1! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/19/2014 15:05'!
pointersPerEntry
	"Values kept in a single entry of a PLC:
	- lookup class
	- method
	- newMethod
	- primitiveIndex"
	
	^ 4! !

!PLCStruct class methodsFor: 'constants' stamp: 'ag 1/22/2014 13:42'!
primitiveIndexOffset
	"Offset within a cache-entry that contains the primitive index of the cached method."
	
	^ 2! !

!PLCStruct class methodsFor: 'statistic' stamp: 'ag 1/21/2014 11:09'!
statistic
	
	^ self statistic:
		#(notNil hasEntries isFull isMonomorphic isPolymorphic isMegamorphic shouldBeMegamorphic)! !

!PLCStruct class methodsFor: 'statistic' stamp: 'ag 1/21/2014 11:08'!
statistic: symbols
	
	^ self allInstances
		inject:
			(symbols collect: [ :symbol | symbol -> 0 ])
		into: [ :statistic :plc |
			statistic do: [ :association |
				(association key value: plc)
					ifTrue: [ association value: association value + 1 ].
			statistic ] ]! !

!PLCStruct class methodsFor: 'statistic' stamp: 'ag 1/22/2014 01:33'!
statisticString: statistic
	
	| stat filled |
	stat := Dictionary newFrom: statistic.
	stat removeKey: #notNil.
	filled := stat at: #hasEntries.
	stat removeKey: #hasEntries.
	^ String streamContents: [ :s |
		stat keysAndValuesDo: [ :sym :num |
			s
				nextPutAll: sym;
				nextPutAll: ': ';
				nextPutAll: (num / 	filled) asFloat asString; cr ] ]! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/30/2014 12:44'!
cacheEntryCounters
	
	^ self entriesAt: self class cacheEntryCounterOffset! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/30/2014 15:33'!
classesAndCounts
	"Return a collection showing all cached classes and their associated
	call frequencies."
	
	| result |
	result := OrderedCollection new.
	self entryIndicesDo:
		[ :i | result add:
			(self at: i + self class lookupClassOffset)
			-> (self at: i + self class cacheEntryCounterOffset) ].
	^ result! !

!PLCStruct methodsFor: 'accessing' stamp: 'ag 1/19/2014 16:56'!
compiledMethod
	
	^ self at: self class compiledMethodIndex! !

!PLCStruct methodsFor: 'accessing' stamp: 'ag 1/19/2014 16:56'!
compiledMethod: method
	
	self at: self class compiledMethodIndex put: method.! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/28/2014 11:10'!
entriesAt: entryOffset
	
	| entries |
	entries := OrderedCollection new.
	self entriesAt: entryOffset do: [ :entry | entries add: entry ].
	^ entries! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/28/2014 11:10'!
entriesAt: entryOffset do: aBlock
	
	^ self entryIndicesAt: entryOffset do: [ :index | aBlock value: (self at: index) ]! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/28/2014 11:09'!
entryIndicesAt: entryOffset do: aBlock
	"entryOffset is one of the *Offset constants on the class side.
	Every *Offset is like an instance variable of a cache entry."
	
	^ self class firstEntryIndex + entryOffset to: self size by: self class pointersPerEntry do: aBlock! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/30/2014 12:44'!
entryIndicesDo: aBlock	
	"Iterate the start-indices of every cache-entry.
	The parts of the entries can be accessed by adding class-side constants to these indices.
	The zero is not a magic constant, it just means we access the beginning of each cache entry."
	
	^ self entryIndicesAt: 0 do: aBlock! !

!PLCStruct methodsFor: 'flushing' stamp: 'ag 1/28/2014 11:23'!
flushAll
	
	self entryIndicesDo: [ :index |
		self invalidateEntry: index ].! !

!PLCStruct methodsFor: 'flushing' stamp: 'ag 1/28/2014 11:23'!
flushMethod: mt
	
	self selector = mt selector ifFalse: [ ^ self ].
	self entryIndicesDo: [ :i |
		mt == (self at: i + self class newMethodOffset)
			ifTrue: [ self invalidateEntry: i ] ].! !

!PLCStruct methodsFor: 'testing' stamp: 'ag 1/30/2014 12:46'!
hasEntries
	
	self
		entriesAt: self class lookupClassOffset
		do: [ :cl | cl ifNotNil: [ ^ true ] ].
	^ false! !

!PLCStruct methodsFor: 'private' stamp: 'ag 1/19/2014 16:48'!
hasLiteral: literal
	"A PLCStruct does not contain any other literals."
	
	^ false! !

!PLCStruct methodsFor: 'flushing' stamp: 'ag 1/30/2014 12:42'!
invalidateEntry: index
	"Flush all cache-entries containing real smalltalk objects.
	Frees them up for GC.
	(Ignore the primitive index and cache entry counter as small optimization. Will be overwritten)"
	
	self
		at: index + self class lookupClassOffset put: nil;
		at: index + self class newMethodOffset put: nil.! !

!PLCStruct methodsFor: 'testing' stamp: 'ag 1/30/2014 12:47'!
isFull
	
	self isMegamorphic ifTrue: [ ^ false ].
	self
		entriesAt: self class lookupClassOffset
		do: [ :cl | cl ifNil: [ ^ false ] ].
	^ true! !

!PLCStruct methodsFor: 'testing' stamp: 'ag 1/21/2014 10:28'!
isMegamorphic
	"Megamorphic PLCs are just 'empty' markers."
	
	^ self size == self class metaInfoEntries! !

!PLCStruct methodsFor: 'testing' stamp: 'ag 1/21/2014 10:30'!
isMonomorphic
	
	^ self numberOfCachedEntries = 1! !

!PLCStruct methodsFor: 'testing' stamp: 'ag 1/21/2014 10:30'!
isPolymorphic
	
	^ self numberOfCachedEntries > 1! !

!PLCStruct methodsFor: 'comparing' stamp: 'ag 1/19/2014 16:10'!
literalEqual: otherObject
	"Every PLCStructs is a different literal, regardless of it's contents."
	
	^ self == otherObject! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/28/2014 11:21'!
methods
	
	^ self entriesAt: self class newMethodOffset! !

!PLCStruct methodsFor: 'accessing' stamp: 'ag 1/30/2014 12:49'!
numberOfCachedEntries
	
	| entries |
	entries := 0.
	self
		entryIndicesAt: self class lookupClassOffset
		do: [ :cl | cl ifNotNil: [ entries := entries + 1 ]].
	^ entries! !

!PLCStruct methodsFor: 'accessing' stamp: 'ag 1/19/2014 15:08'!
numberOfEntries
	
	^ (self size - self class metaInfoEntries) // self class pointersPerEntry! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/28/2014 11:22'!
primitiveIndices
	
	^ self entriesAt: self class primitiveIndexOffset! !

!PLCStruct methodsFor: 'accessing cache entries' stamp: 'ag 1/28/2014 11:12'!
receiverClasses
	
	^ self entriesAt: self class lookupClassOffset! !

!PLCStruct methodsFor: 'accessing' stamp: 'ag 1/30/2014 13:06'!
resetCacheEntryCounters
	
	self
		entryIndicesAt: self class cacheEntryCounterOffset
		do: [ :i | self at: i put: 0 ]! !

!PLCStruct methodsFor: 'accessing' stamp: 'ag 1/19/2014 16:56'!
selector
	
	^ self at: self class messageSelectorIndex! !

!PLCStruct methodsFor: 'accessing' stamp: 'ag 1/19/2014 16:56'!
selector: selector
	
	self at: self class messageSelectorIndex put: selector.! !

!PLCStruct methodsFor: 'testing' stamp: 'ag 1/21/2014 10:30'!
shouldBeMegamorphic
	
	^ self numberOfCachedEntries >= 10! !

!InstructionStream methodsFor: '*VMs-Seminar' stamp: 'ag 1/16/2014 15:08'!
aboutToSend
	"Skip bytes reserved for plc literal index."
	
	self method flag ifTrue: [ pc := pc + 2 ].! !

!InstructionStream methodsFor: '*VMs-Seminar' stamp: 'ag 1/16/2014 15:08'!
interpretExtension: offset in: method for: client
	| type offset2 byte2 byte3 byte4 |
	offset <= 6 ifTrue: 
		["Extended op codes 128-134"
		byte2 := method at: pc. pc := pc + 1.
		offset <= 2 ifTrue:
			["128-130:  extended pushes and pops"
			type := byte2 // 64.
			offset2 := byte2 \\ 64.
			offset = 0 ifTrue: 
				[type = 0 ifTrue: [^client pushReceiverVariable: offset2].
				type = 1 ifTrue: [^client pushTemporaryVariable: offset2].
				type = 2  ifTrue: [^client pushConstant: (method literalAt: offset2 + 1)].
				type = 3 ifTrue: [^client pushLiteralVariable: (method literalAt: offset2 + 1)]].
			offset = 1 ifTrue: 
				[type = 0 ifTrue: [^client storeIntoReceiverVariable: offset2].
				type = 1 ifTrue: [^client storeIntoTemporaryVariable: offset2].
				type = 2 ifTrue: [self error: 'illegalStore'].
				type = 3 ifTrue: [^client storeIntoLiteralVariable: (method literalAt: offset2 + 1)]].
			offset = 2 ifTrue: 
				[type = 0 ifTrue: [^client popIntoReceiverVariable: offset2].
				type = 1 ifTrue: [^client popIntoTemporaryVariable: offset2].
				type = 2 ifTrue: [self error: 'illegalStore'].
				type = 3  ifTrue: [^client popIntoLiteralVariable: (method literalAt: offset2 + 1)]]].
		"131-134: extended sends"
		offset = 3 ifTrue:  "Single extended send"
			[
			self aboutToSend.
			^client send: (method literalAt: byte2 \\ 32 + 1)
					super: false numArgs: byte2 // 32].
		offset = 4 ifTrue:    "Double extended do-anything"
			[byte3 := method at: pc. pc := pc + 1.
			type := byte2 // 32.
			type = 0 ifTrue: [
				self aboutToSend.
				^client send: (method literalAt: byte3 + 1)
									super: false numArgs: byte2 \\ 32].
			type = 1 ifTrue: [
				self aboutToSend.
				^client send: (method literalAt: byte3 + 1)
									super: true numArgs: byte2 \\ 32].
			type = 2 ifTrue: [^client pushReceiverVariable: byte3].
			type = 3 ifTrue: [^client pushConstant: (method literalAt: byte3 + 1)].
			type = 4 ifTrue: [^client pushLiteralVariable: (method literalAt: byte3 + 1)].
			type = 5 ifTrue: [^client storeIntoReceiverVariable: byte3].
			type = 6 ifTrue: [^client popIntoReceiverVariable: byte3].
			type = 7 ifTrue: [^client storeIntoLiteralVariable: (method literalAt: byte3 + 1)]].
		offset = 5 ifTrue:  "Single extended send to super"
			[
			self aboutToSend.
			^client send: (method literalAt: byte2 \\ 32 + 1)
					super: true numArgs: byte2 // 32].
		offset = 6 ifTrue:   "Second extended send"
			[
			self aboutToSend.
			^client send: (method literalAt: byte2 \\ 64 + 1)
					super: false numArgs: byte2 // 64]].
	offset = 7 ifTrue: [^client doPop].
	offset = 8 ifTrue: [^client doDup].
	offset = 9 ifTrue: [^client pushActiveContext].
	byte2 := method at: pc. pc := pc + 1.
	offset = 10 ifTrue:
		[^byte2 < 128
			ifTrue: [client pushNewArrayOfSize: byte2]
			ifFalse: [client pushConsArrayWithElements: byte2 - 128]].
	offset = 11 ifTrue: [^self error: 'unusedBytecode'].
	byte3 := method at: pc.  pc := pc + 1.
	offset = 12 ifTrue: [^client pushRemoteTemp: byte2 inVectorAt: byte3].
	offset = 13 ifTrue: [^client storeIntoRemoteTemp: byte2 inVectorAt: byte3].
	offset = 14 ifTrue: [^client popIntoRemoteTemp: byte2 inVectorAt: byte3].
	"offset = 15"
	byte4 := method at: pc.  pc := pc + 1.
	^client
		pushClosureCopyNumCopiedValues: (byte2 bitShift: -4)
		numArgs: (byte2 bitAnd: 16rF)
		blockSize: (byte3 * 256) + byte4! !

!InstructionStream methodsFor: '*VMs-Seminar' stamp: 'ag 1/16/2014 15:09'!
interpretNextInstructionFor: client 
	"Send to the argument, client, a message that specifies the type of the 
	next instruction."

	| byte type offset method |
	method := self method.  
	byte := method at: pc.
	type := byte // 16.  
	offset := byte \\ 16.  
	pc := pc+1.
	"We do an inline binary search on each of the possible 16 values of type:
	The old, cleaner but slowe code is retained as a comment below"
	type < 8
	ifTrue: [type < 4
				ifTrue: [type < 2
						ifTrue: [type < 1
								ifTrue: ["type = 0"
									^ client pushReceiverVariable: offset]
								ifFalse: ["type = 1"
									^ client pushTemporaryVariable: offset]]
						ifFalse: [type < 3
								ifTrue: ["type = 2"
									^ client
										pushConstant: (method literalAt: offset + 1)]
								ifFalse: ["type = 3"
									^ client
										pushConstant: (method literalAt: offset + 17)]]]
				ifFalse: [type < 6
						ifTrue: [type < 5
								ifTrue: ["type = 4"
									^ client
										pushLiteralVariable: (method literalAt: offset + 1)]
								ifFalse: ["type = 5"
									^ client
										pushLiteralVariable: (method literalAt: offset + 17)]]
						ifFalse: [type < 7
								ifTrue: ["type = 6"
									offset < 8
										ifTrue: [^ client popIntoReceiverVariable: offset]
										ifFalse: [^ client popIntoTemporaryVariable: offset - 8]]
								ifFalse: ["type = 7"
									offset = 0
										ifTrue: [^ client pushReceiver].
									offset < 8
										ifTrue: [^ client
												pushConstant: (SpecialConstants at: offset)].
									offset = 8
										ifTrue: [^ client methodReturnReceiver].
									offset < 12
										ifTrue: [^ client
												methodReturnConstant: (SpecialConstants at: offset - 8)].
									offset = 12
										ifTrue: [^ client methodReturnTop].
									offset = 13
										ifTrue: [^ client blockReturnTop].
									offset = 14
										ifTrue: [
											"actually an unused bytecode, but misused as 'very long jump'.
											Added for VMs-Seminar."
											| byte1 byte2 |
											byte1 := method at: pc.  pc:= pc+1.
											byte2 := method at: pc.  pc:= pc+1.
											^ client jump: (byte1 - 128) * 256 + byte2 ].
									offset > 13
										ifTrue: [^ self error: 'unusedBytecode']]]]]
		ifFalse: [type < 12
				ifTrue: [type < 10
						ifTrue: [type < 9
								ifTrue: ["type = 8"
									^ self
										interpretExtension: offset
										in: method
										for: client]
								ifFalse: ["type = 9 (short jumps)"
									offset < 8
										ifTrue: [^ client jump: offset + 1].
									^ client jump: offset - 8 + 1 if: false]]
						ifFalse: [type < 11
								ifTrue: ["type = 10 (long jumps)"
									byte := method at: pc.
									pc := pc + 1.
									offset < 8
										ifTrue: [^ client jump: offset - 4 * 256 + byte].
									^ client jump: (offset bitAnd: 3)
											* 256 + byte if: offset < 12]
								ifFalse: ["type = 11"
									self aboutToSend.
									^ client
										send: (Smalltalk specialSelectorAt: offset + 1)
										super: false
										numArgs: (Smalltalk specialNargsAt: offset + 1)]]]
				ifFalse: [type = 12
						ifTrue: [
							self aboutToSend.
							^ client
								send: (Smalltalk specialSelectorAt: offset + 17)
								super: false
								numArgs: (Smalltalk specialNargsAt: offset + 17)]
						ifFalse: ["type = 13, 14 or 15"
							self aboutToSend.
							^ client
								send: (method literalAt: offset + 1)
								super: false
								numArgs: type - 13]]].


"    old code 
	type=0 ifTrue: [^client pushReceiverVariable: offset].
	type=1 ifTrue: [^client pushTemporaryVariable: offset].
	type=2 ifTrue: [^client pushConstant: (method literalAt: offset+1)].
	type=3 ifTrue: [^client pushConstant: (method literalAt: offset+17)].
	type=4 ifTrue: [^client pushLiteralVariable: (method literalAt: offset+1)].
	type=5 ifTrue: [^client pushLiteralVariable: (method literalAt: offset+17)].
	type=6 
		ifTrue: [offset<8
					ifTrue: [^client popIntoReceiverVariable: offset]
					ifFalse: [^client popIntoTemporaryVariable: offset-8]].
	type=7
		ifTrue: [offset=0 ifTrue: [^client pushReceiver].
				offset<8 ifTrue: [^client pushConstant: (SpecialConstants at: offset)].
				offset=8 ifTrue: [^client methodReturnReceiver].
				offset<12 ifTrue: [^client methodReturnConstant: 
												(SpecialConstants at: offset-8)].
				offset=12 ifTrue: [^client methodReturnTop].
				offset=13 ifTrue: [^client blockReturnTop].
				offset>13 ifTrue: [^self error: 'unusedBytecode']].
	type=8 ifTrue: [^self interpretExtension: offset in: method for: client].
	type=9
		ifTrue:  short jumps
			[offset<8 ifTrue: [^client jump: offset+1].
			^client jump: offset-8+1 if: false].
	type=10 
		ifTrue:  long jumps
			[byte:= method at: pc.  pc:= pc+1.
			offset<8 ifTrue: [^client jump: offset-4*256 + byte].
			^client jump: (offset bitAnd: 3)*256 + byte if: offset<12].
	type=11 
		ifTrue: 
			[^client 
				send: (Smalltalk specialSelectorAt: offset+1) 
				super: false
				numArgs: (Smalltalk specialNargsAt: offset+1)].
	type=12 
		ifTrue: 
			[^client 
				send: (Smalltalk specialSelectorAt: offset+17) 
				super: false
				numArgs: (Smalltalk specialNargsAt: offset+17)].
	type>12
		ifTrue: 
			[^client send: (method literalAt: offset+1) 
					super: false
					numArgs: type-13]"! !

!InstructionStream methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 13:04'!
nextPc: currentByte
	"Answer the pc of the next bytecode following the current one, given the current bytecode.."

	| type |
	currentByte = 126 ifTrue: [ ^ pc + 3 ]. "Very long jump"
	type := currentByte // 16.
	^type = 8 "extensions"
				ifTrue: [pc + (#(2 2 2 2 3 2 2 1 1 1 2 1 3 3 3 4) at: currentByte \\ 16 + 1)]
				ifFalse: [type = 10 "long jumps"
							ifTrue: [pc + 2]
							ifFalse: [pc + 1]]! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/15/2014 10:22'!
allLocals
	"The merge of locals and declarations (don't ask)"

	^(declarations keys asSet reject: [:k| (declarations at: k) == #implicit])
		addAll: locals; yourself! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:49'!
comment
	^comment! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:50'!
elideAnyFinalReturn
	"For super expansions we need to eliminate any final return to prevent premature exit.
	 Anything meaningful in the returned expression must be retained."

	| stmtList expr |
	stmtList := parseTree statements asOrderedCollection.
	stmtList last isReturn ifTrue:
		[expr := stmtList last expression.
		 (expr isVariable and: [expr name = 'self'])
			ifTrue: [stmtList := stmtList allButLast]
			ifFalse: [stmtList at: stmtList size put: expr].
		parseTree setStatements: stmtList]! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:50'!
extraVariableNumber
	^extraVariableNumber! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:47'!
hasProperties
	^properties notNil and: [properties notEmpty]! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:45'!
mergePropertiesOfSuperMethod: superTMethod
	superTMethod hasProperties ifFalse:
		[^self].
	self hasProperties ifFalse:
		[properties := superTMethod properties.
		 ^self].
	superTMethod properties pragmas do:
		[:aPragma|
		(self shouldIncorporatePragmaFromSuperMethod: aPragma) ifTrue:
			[properties := properties copyWith: aPragma]]! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:47'!
properties
	^ properties! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:46'!
shouldIncorporatePragmaFromSuperMethod: aPragma
	(properties includesKey: aPragma keyword) ifFalse:
		[^true].
	((aPragma keyword beginsWith: #var:)
	  and: [properties pragmas noneSatisfy:
			[:p|
			(p keyword beginsWith: #var:)
			and: [(p argumentAt: 1) = (aPragma argumentAt: 1)]]]) ifTrue:
		[^true].
	^false! !

!TMethod methodsFor: '*VMs-Seminar' stamp: 'ag 1/14/2014 13:44'!
superExpansionNodeFor: aSelector args: argumentNodes
	"Answer the expansion of a super send.  Merge the super expansion's
	 locals, properties and comment into this method's properties."
	
	(definingClass superclass lookupSelector: aSelector)
		ifNil: [self error: 'superclass does not define super method']
		ifNotNil:
			[:superMethod| | superTMethod commonVars varMap |
			superTMethod := superMethod methodNode asTranslationMethodOfClass: self class.
			((argumentNodes allSatisfy: [:parseNode| parseNode isVariableNode])
			and: [(argumentNodes asOrderedCollection collect: [:parseNode| parseNode key]) = superTMethod args]) ifFalse:
				[self error: definingClass name, '>>',selector, ' args ~= ',
							superTMethod definingClass name, '>>', aSelector,
							(String with: $. with: Character cr),
							'For super expansions to be translated correctly each argument must be a variable with the same name as the corresponding argument in the super method.'].
			self mergePropertiesOfSuperMethod: superTMethod.
			(commonVars := superTMethod locals intersection: self locals) notEmpty ifTrue:
				[varMap := Dictionary new.
				 commonVars do:
					[:k| varMap at: k put: (superTMethod unusedNamePrefixedBy: k avoiding: self allLocals)].
				 superTMethod renameVariablesUsing: varMap].
			self assert: (superTMethod locals allSatisfy: [:var| (self locals includes: var) not]).
			locals addAll: superTMethod locals.
			superTMethod declarations keysAndValuesDo:
				[:var :decl|
				self declarationAt: var put: decl].
			superTMethod comment ifNotNil:
				[:superComment|
				comment := comment
								ifNil: [superComment]
								ifNotNil: [superComment, comment]].
			superTMethod extraVariableNumber ifNotNil:
				[:scvn|
				extraVariableNumber := extraVariableNumber ifNil: [scvn] ifNotNil: [:cvn| cvn + scvn]].
			superTMethod elideAnyFinalReturn.
			^superTMethod parseTree]! !

!MessageNode methodsFor: '*VMs-Seminar' stamp: 'ag 1/22/2014 15:30'!
sizeCodeForToDo: encoder value: forValue 
	" var := rcvr. L1: [var <= arg1] Bfp(L2) [block body. var := var + inc] Jmp(L1) L2: "
	| loopSize initStmt test block incStmt blockSize initSize limitInit |
	block := arguments at: 3.
	initStmt := arguments at: 4.
	test := arguments at: 5.
	incStmt := arguments at: 6.
	limitInit := arguments at: 7.
	initSize := initStmt sizeCodeForEffect: encoder.
	limitInit == nil ifFalse:
		[initSize := initSize + (limitInit sizeCodeForEffect: encoder)].
	blockSize := (block sizeCodeForEvaluatedEffect: encoder)
			+ (incStmt sizeCodeForEffect: encoder).
	blockSize := blockSize + (encoder sizeJumpLong: blockSize negated).
	loopSize := (test sizeCodeForValue: encoder)
			+ (self sizeCode: encoder forBranchOn: false dist: blockSize)
			+ blockSize.
	sizes := Array with: blockSize with: loopSize.
	^initSize
	+ loopSize
	+ (forValue ifTrue: [encoder sizePushSpecialLiteral: nil] ifFalse: [0])! !

Interpreter subclass: #VeryLongJumpInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar'!

VeryLongJumpInterpreter subclass: #NoMethodCacheInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMs-Seminar'!

!NoMethodCacheInterpreter commentStamp: '<historical>' prior: 0!
This class is a complete implementation of the Smalltalk-80 virtual machine, derived originally from the Blue Book specification but quite different in some areas.

It has been modernized with 32-bit pointers, better management of Contexts, and attention to variable use that allows the CCodeGenerator (qv) to keep, eg, the instruction pointer and stack pointer in registers as well as keeping most simple variables in a global array that seems to improve performance for most platforms.

In addition to SmallInteger arithmetic and Floats, it supports logic on 32-bit PositiveLargeIntegers, thus allowing it to simulate itself much more effectively than would otherwise be the case.

NOTE:  Here follows a list of things to be borne in mind when working on this code, or when making changes for the future.

1.  There are a number of things that should be done the next time we plan to release a copletely incompatible image format.  These include unifying the instanceSize field of the class format word -- see instantiateClass:indexableSize:, and unifying the bits of the method primitive index (if we decide we need more than 512, after all) -- see primitiveIndexOf:.  Also, contexts should be given a special format code (see next item).

2.  There are several fast checks for contexts (see isContextHeader: and isMethodContextHeader:) which will fail if the compact class indices of BlockContext or MethodContext change.  This is necessary because the oops may change during a compaction when the oops are being adjusted.  It's important to be aware of this when writing a new image using the systemTracer.  A better solution would be to reserve one of the format codes for Contexts only.

3.  We have made normal files tolerant to size and positions up to 32 bits.  This has not been done for async files, since they are still experimental.  The code in size, at: and at:put: should work with sizes and indices up to 31 bits, although I have not tested it (di 12/98); it might or might not work with 32-bit sizes.

4.  Note that 0 is used in a couple of places as an impossible oop.  This should be changed to a constant that really is impossible (or perhaps there is code somewhere that guarantees it --if so it should be put in this comment).  The places include the method cache and the at cache. !

!NoMethodCacheInterpreter methodsFor: 'method lookup cache' stamp: 'ag 1/19/2014 17:04'!
addNewMethodToCache
	"Method cache disabled."
	
	<inline: false>
	self compilerTranslateMethodHook.	"newMethod x lkupClass -> newNativeMethod (may cause GC !!)"
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: lkupClass.! !

!NoMethodCacheInterpreter methodsFor: 'method lookup cache' stamp: 'ag 1/19/2014 17:04'!
lookupInMethodCacheSel: selector class: class
	"Lookup Cache disabled."
	
	<inline>
	^ false! !

VeryLongJumpInterpreter subclass: #PLCInterpreter
	instanceVariableNames: 'syntheticMessageSend'
	classVariableNames: 'DebugPLC PlcCacheEntryCounterOffset PlcCompiledMethodIndex PlcEntries PlcFirstEntryIndex PlcIncrementCacheEntryCounters PlcLkupClassIndex PlcMessageSelectorIndex PlcMetaInfoEntries PlcNewMethodIndex PlcPrimitiveIndex PlcUseMethodLookupCache PointersPerPlcEntry TracePLC'
	poolDictionaries: ''
	category: 'VMs-Seminar'!

!PLCInterpreter commentStamp: '<historical>' prior: 0!
This class is a complete implementation of the Smalltalk-80 virtual machine, derived originally from the Blue Book specification but quite different in some areas.

It has been modernized with 32-bit pointers, better management of Contexts, and attention to variable use that allows the CCodeGenerator (qv) to keep, eg, the instruction pointer and stack pointer in registers as well as keeping most simple variables in a global array that seems to improve performance for most platforms.

In addition to SmallInteger arithmetic and Floats, it supports logic on 32-bit PositiveLargeIntegers, thus allowing it to simulate itself much more effectively than would otherwise be the case.

NOTE:  Here follows a list of things to be borne in mind when working on this code, or when making changes for the future.

1.  There are a number of things that should be done the next time we plan to release a copletely incompatible image format.  These include unifying the instanceSize field of the class format word -- see instantiateClass:indexableSize:, and unifying the bits of the method primitive index (if we decide we need more than 512, after all

----STARTUP----{10 December 2013 . 3:43:28 pm} as /home/anton/shared/squeak/Squeak4.3.image!

!PLCInterpreter class methodsFor: 'initialization' stamp: 'ag 1/15/2014 09:49'!
initialize
	
	super initialize.
	self initializePLC.! !

!PLCInterpreter class methodsFor: 'translation' stamp: 'ag 1/15/2014 09:54'!
initializeClassicObjectMemoryInCodeGenerator: cg
	"Copied from initializeClassicObjectMemoryInCodeGenerator:
	This is hacky, but saw no other way without adding ObjectMemory multiple times."
	
	cg addClass: PLCObjectMemory upTo: VMClass asInstanceVariable: #objectMemory.
	"references from ObjectMemory to #interpreter are local functions in generated code"
	^cg mapVar: #interpreter asInstanceOf: PLCObjectMemory to: 'self'! !

!PLCInterpreter class methodsFor: 'initialization' stamp: 'ag 2/1/2014 13:10'!
initializePLC
	
	DebugPLC := false.
	TracePLC := false.
	PlcUseMethodLookupCache := true.
	PlcIncrementCacheEntryCounters := false.
	
	PlcEntries := PLCStruct defaultNumberOfEntries.
	PointersPerPlcEntry := PLCStruct pointersPerEntry.
	
	PlcMetaInfoEntries := PLCStruct metaInfoEntries.
	PlcFirstEntryIndex := PlcMetaInfoEntries.
	PlcMessageSelectorIndex := PLCStruct messageSelectorIndex - 1.
	PlcCompiledMethodIndex := PLCStruct compiledMethodIndex - 1.
	
	"These should all be called *Offset"
	PlcLkupClassIndex := PLCStruct lookupClassOffset.
	PlcNewMethodIndex := PLCStruct newMethodOffset.
	PlcPrimitiveIndex := PLCStruct primitiveIndexOffset.
	PlcCacheEntryCounterOffset := PLCStruct cacheEntryCounterOffset.
	
	PLCObjectMemory initializeSpecialObjectIndices.! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/21/2014 10:39'!
addToPLC: plc
	"Linear search over the contents of the PLCStruct until an empty spot is found.
	Return false if the PLCStruct was already full.
	C compiler should unroll constant loop."
	
	"TODO - Handle the already-full condition by invalidating a value or growing the PLC."
	
	<inline: true>
	| index |
	index := self indexOfClass: objectMemory nilObj inPLC: plc.
	index = 0 ifTrue: [
		self debugFullPLC: plc.
		^ false ].
	self storeIntoPlc: plc index: index.
	^ true! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:43'!
checkOkayOop: oop
	
	<inline: false>
	objectMemory okayOop: oop.
	self oopHasOkayClass: oop.! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/20/2014 18:48'!
checkString: stString equals: cString
	"Mostly copied from #classNameOf:Is:"
	
	<inline: false>
	<var: #cString type: 'char *'>
	<var: #arr type: 'char *'>
	| length arr |
	(objectMemory isBytes: stString) ifFalse: [^ false].
	length := self stSizeOf: stString.
	arr := self cCoerce: (self arrayValueOf: stString) to: 'char *'.
	0 to: length - 1 do: [:i | (cString at: i) = (arr at: i) ifFalse: [^ false]].
	"Check if cString really ends at this point"
	^ (cString at: length) = 0! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/18/2014 00:06'!
classNameOf: aClass
	
	^ objectMemory fetchPointer: 6 ofObject: aClass! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/19/2014 15:03'!
createNewPLCAt: plcIndex
	
	| plc |
	plc := self instantiatePLC.
	plc storePointer: PlcMessageSelectorIndex ofObject: plc withValue: messageSelector.
	plc storePointer: PlcCompiledMethodIndex ofObject: plc withValue: method.
	self literal: plcIndex ofMethod: method store: plc.
	self storeIntoPlc: plc index: PlcFirstEntryIndex. "Offset for storing selector"! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:44'!
debugCorrectClassOf: plc at: plcIndex
	
	self isInDebuggedMethod ifFalse: [ ^ true ].
	(objectMemory fetchClassOf: plc) == objectMemory classPLCStruct
		ifFalse: [
			self print: 'Wrong class of plc object for '.
			self printStringOf: messageSelector.
			self print: ' at index '.
			self printNum: plcIndex.
			self print: ' inside method '.
			self printStringOf: (self selectorOfMethod: method).
			self cr.
			^ false ].
	^ true! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:42'!
debugFullPLC: plc
	
	<inline: false>
	self isInDebuggedMethod ifFalse: [ ^ nil ].
	self print: 'PLC is full!! '.
	self printPLC: plc.
	self cr.! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:44'!
debugOkayPLC: plc
	
	<inline: false>
	(DebugPLC and: [ self isInDebuggedMethod ]) ifFalse: [ ^ nil ].
	self print: ' > Checking consistency of PLC entries .. '.
	0 to: self plcSize - 1 do: [ :i |
		self printNum: i. self print: ' .. '.
		objectMemory checkOkayOop: (objectMemory fetchPointer: i ofObject: plc)
	].
	self cr.! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:42'!
debugPLC: plc atIndex: index
	
	<inline: false>
	self isInDebuggedMethod ifFalse: [ ^ nil ].
	self printPLC: plc atIndex: index.
	self cr.! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:42'!
debugRestoredFrom: plc index: index
	
	<inline: false>
	self isInDebuggedMethod ifFalse: [ ^ nil ].
	self print: ' > Restored from PLC. '.
	self printCacheEntry: index of: plc.
	self cr.! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:42'!
debugStoredInto: plc index: index
	
	<inline: false>
	self isInDebuggedMethod ifFalse: [ ^ nil ].
	self print: ' > Stored into PLC. '.
	self printCacheEntry: index of: plc.
	self cr.! !

!PLCInterpreter methodsFor: 'message sending' stamp: 'ag 1/20/2014 17:22'!
doMethodLookup
	"Small configurable optimization: instead of doing just the regular lookup, use the method lookup cache.
	If disabled, just do the regular method lookup."
	
	PlcUseMethodLookupCache
		ifTrue: [	super internalFindNewMethod ]
		ifFalse: [
			self externalizeIPandSP.
			self lookupMethodInClass: lkupClass.
			self internalizeIPandSP.
			self compilerTranslateMethodHook.
			self setPrimitiveFunctionPointer: primitiveIndex ].! !

!PLCInterpreter methodsFor: 'compiled methods' stamp: 'ag 1/13/2014 17:52'!
flagBitOf: aMethod
	
	<inline: true>
	^ ((self headerOf: aMethod) >> 30) bitAnd: 16r1! !

!PLCInterpreter methodsFor: 'message sending' stamp: 'ag 1/21/2014 00:11'!
handlingPLCMethod
	"Three conditions for using the Polymorphic Lookup Cache for the current message-send:
	1. The current message-send is not generate by the VM. There are no PLCStructs at these synthetic call-sites.
	2. The VM has found the class PLCStruct in the specialObjectsArray
	3. The current CompiledMethod has the flag-big set in it's header, indicating that it contains PLC call-sites."
	
	<inline: true>
	^ (syntheticMessageSend not
	and: [ objectMemory hasPLCStruct ])
	and: [ self flagBitOf: method ]! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 2/1/2014 13:11'!
incrementCacheEntryCounter: index of: plc
	
	<inline: true>
	| counterPointer counter |
	PlcIncrementCacheEntryCounters ifFalse: [ ^ nil ].
	counterPointer := objectMemory fetchPointer: index + PlcCacheEntryCounterOffset ofObject: plc.
	(objectMemory isIntegerObject: counterPointer ) ifFalse: [ ^ nil ].
	counter := objectMemory integerValueOf: counterPointer.
	counter := counter + 1.
	objectMemory
		storePointer: index + PlcCacheEntryCounterOffset
		ofObject: plc
		withValue: (objectMemory integerObjectOf: counter).! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/19/2014 15:04'!
indexOfClass: class inPLC: plc
	"Iterate the class-slots in plc and return the index of class, if found. Return 0 otherwise.
	0 cannot be the index of any class, because the selector is stored at index 0."
	
	PlcFirstEntryIndex to: self plcSize - 1 by: PointersPerPlcEntry do: [ :index |
		class == (objectMemory fetchPointer: index + PlcLkupClassIndex ofObject: plc)
			ifTrue: [ ^ index ] ].
	^ 0! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/17/2014 23:54'!
instantiateClass: class ofSize: size
	
	| spaceOkay |
	spaceOkay := self sufficientSpaceToInstantiate: class indexableSize: size.
	self success: spaceOkay.
	^ objectMemory instantiateClass: class indexableSize: size! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/17/2014 23:54'!
instantiatePLC
	"Return an uninitialized instance of PLCStruct.
	This assumes that 'objectMemory hasPLCStruct' returns true!! Otherwise probably segfault."
	
	^ self instantiateClass: ObjectMemory classPLCStruct ofSize: self plcSize! !

!PLCInterpreter methodsFor: 'return bytecodes' stamp: 'ag 1/20/2014 23:59'!
internalAboutToReturn: resultObj through: aContext
	<inline: true>
	syntheticMessageSend := true.
	^ super internalAboutToReturn: resultObj through: aContext! !

!PLCInterpreter methodsFor: 'return bytecodes' stamp: 'ag 1/20/2014 23:59'!
internalCannotReturn: resultObj
	<inline: true>
	syntheticMessageSend := true.
	^ super internalCannotReturn: resultObj! !

!PLCInterpreter methodsFor: 'message sending' stamp: 'ag 1/21/2014 10:58'!
internalFindNewMethod
	"If current method is not flagged, fall back to regular method lookup cache.
	Otherwise load the PLC data at current callsite and handle the PLC.
	Afterwards, jump over the 2 bytes after the current send-bytecode."
	
	<inline: true>
	| handlingPLC plcIndex plc |
	handlingPLC := self handlingPLCMethod.
	syntheticMessageSend := false.
	handlingPLC ifFalse: [ ^ self doMethodLookup ].
	plcIndex := self plcIndexAtCallsite: localIP.
	plc := self literal: plcIndex.
	(self debugCorrectClassOf: plc at: plcIndex) ifFalse: [ ^ self doMethodLookup. ].
	self internalHandlePLC: plc at: plcIndex.
	self internalJumpOverCallsite.! !

!PLCInterpreter methodsFor: 'message sending' stamp: 'ag 1/30/2014 12:34'!
internalHandlePLC: plc at: plcIndex
	"If a PLC is already present, do a lookup.
	If no PLC is present, or the lookup fails, do a regular lookup.
	Then create a new PLC if required and add the new entry."
	
	self debugPLC: plc atIndex: plcIndex.
	(self lookupClass: lkupClass inPLC: plc )
		ifFalse: [
			self doMethodLookup.
			self addToPLC: plc ].! !

!PLCInterpreter methodsFor: 'message sending' stamp: 'ag 1/16/2014 15:26'!
internalJumpOverCallsite
	"Skip the two bytes that are located right after a send-bytecode."
	
	<inline: true>
	localIP := localIP + 2.! !

!PLCInterpreter methodsFor: 'compiled methods' stamp: 'ag 1/15/2014 11:01'!
isBytesObject: oop
	
	^ (objectMemory formatOf: oop) >= 8! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/21/2014 10:41'!
isInDebuggedMethod
	"Insert selectors of interest in the strings below."
	
	^ 
	TracePLC or: [
		(DebugPLC
		and: [ (self checkString: messageSelector equals: 'changeRecordForOverriddenMethod:') ]) "invoked method."
		and: [ self checkString: (self selectorOfMethod: method) equals: 'isOverrideOfYourMethod:' ] ] "containing method"! !

!PLCInterpreter methodsFor: 'jump bytecodes' stamp: 'ag 1/20/2014 23:59'!
jumplfFalseBy: offset
	<inline: true>
	syntheticMessageSend := true.
	^ super jumplfFalseBy: offset! !

!PLCInterpreter methodsFor: 'jump bytecodes' stamp: 'ag 1/20/2014 23:59'!
jumplfTrueBy: offset
	<inline: true>
	syntheticMessageSend := true.
	^ super jumplfTrueBy: offset! !

!PLCInterpreter methodsFor: 'compiled methods' stamp: 'ag 1/16/2014 15:55'!
literal: offset ofMethod: methodPointer store: value
	
	objectMemory storePointer: offset + LiteralStart ofObject: methodPointer withValue: value.! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/17/2014 20:36'!
lookupClass: class inPLC: plc
	"Linear search over the contents of the PLCStruct until we find the required class.
	Return false if the class is not find anywhere.
	C compiler should unroll constant loop."
	
	<inline: true>
	| index |
	index := self indexOfClass: class inPLC: plc.
	index = 0 ifTrue: [ ^ false ].
	self restoreFromPlc: plc index: index.
	^ true! !

!PLCInterpreter methodsFor: 'message sending' stamp: 'ag 1/17/2014 20:41'!
plcIndexAtCallsite: localInstructionPointer
	"After a send-bytecode, we insert an ignored jump+1 instruction,
	and after that we have the literal-index for the plc of the current callsite."
	
	<inline: true>
	^ objectMemory byteAtPointer: localInstructionPointer + 2! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/19/2014 15:00'!
plcSize
	"We add one pointer to store the selector associated with the PLCStruct."
	
	^ PlcMetaInfoEntries + (PlcEntries * PointersPerPlcEntry)! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/17/2014 18:47'!
printBits: b
	"b has 4 bytes.
	Print bits, starting with most significant."
	
	0 to: 31 do: [ :i |
		((b bitAnd: (1 << (31 - i))) == 0)
			ifTrue: [ self print: '0' ]
			ifFalse: [ self print: '1' ].
		((i+1) \\ 8) == 0 ifTrue: [ self print: ' ' ].
	]! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/30/2014 13:11'!
printCacheEntry: index of: plc
	
	self print: 'Cache-Entry index '.
	self printNum: index.
	self print: ' - Class: '.
	self printClass: (objectMemory fetchPointer: index + PlcLkupClassIndex ofObject: plc).
	self print: ' NewMethod: '.
	self printPointer: (objectMemory fetchPointer: index + PlcNewMethodIndex ofObject: plc).
	self print: ' PrimitiveIndex: '.
	self printPointer: (objectMemory fetchPointer: index + PlcPrimitiveIndex ofObject: plc).
	self print: ' CacheEntryCounter: '.
	self printPointer: (objectMemory fetchPointer: index + PlcCacheEntryCounterOffset ofObject: plc).! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/18/2014 00:09'!
printClass: class
	
	self printPointer: class.
	self print: ' ('.
	self printStringOf: (self classNameOf: class).
	self print: ')'.! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/30/2014 12:35'!
printPLC: plc
	
	self printPointer: plc.
	self print: ' <'.
	self printStringOf: (objectMemory fetchPointer: PlcMessageSelectorIndex ofObject: plc).
	self print: '> '.
	PlcFirstEntryIndex to: self plcSize - 1 by: PointersPerPlcEntry do: [ :i | | class |
		self print: '['.
		class := objectMemory fetchPointer: i ofObject: plc.
		class = objectMemory nilObj
			ifTrue: [ self print: '<nil>' ]
			ifFalse: [ self printClass: class ].
		self print: '] ' ].! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/18/2014 00:22'!
printPLC: plc atIndex: index
	
	self print: 'PLC index: '.
	self printNum: index.
	self print: ', '.
	self printPLC: plc.! !

!PLCInterpreter methodsFor: 'debugging' stamp: 'ag 1/17/2014 18:52'!
printPointer: p
	
	self cCode: 'printf("%p", (void*) p)'.! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/30/2014 12:59'!
restoreFromPlc: plc index: index
	"Index is already a valid index into the PLCStruct.
	Using localPrimIndex is a small optimization to avoid some indirect memory accesses."
	
	newMethod := objectMemory fetchPointer: index + PlcNewMethodIndex ofObject: plc.
	primitiveIndex := objectMemory integerValueOf: (objectMemory fetchPointer: index + PlcPrimitiveIndex ofObject: plc).
	self incrementCacheEntryCounter: index of: plc.
	self setPrimitiveFunctionPointer: primitiveIndex.
	self debugRestoredFrom: plc index: index.! !

!PLCInterpreter methodsFor: 'compiled methods' stamp: 'ag 1/15/2014 11:03'!
selectorOfMethod: mt
	"The second last literal of a method usually either contains the selector as a symbol, or
	an instance of AdditionalMethodState that has the selector as instance variable."
	
	| literal |
	literal := self literal: (self literalCountOf: mt) - 2 ofMethod: mt.
	(self isBytesObject: literal) ifTrue: [ ^ literal ].
	^ self fetchPointer: 1 ofObject: literal! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/30/2014 12:37'!
setPrimitiveFunctionPointer: index
	
	index = 0 ifFalse: [
		primitiveFunctionPointer := self functionPointerFor: index inClass: lkupClass ].! !

!PLCInterpreter methodsFor: 'polymorphic lookup cache' stamp: 'ag 1/30/2014 13:03'!
storeIntoPlc: plc index: index
	"Put the current state into the plc at the given index.
	Optimization - do the young-check once and store pointers unchecked. See #storePointer:ofObject:withValue:"
	
	(objectMemory isOldNonIntegerOop: plc) ifTrue: [
		self possibleRootStoreInto: plc value: lkupClass.
		self possibleRootStoreInto: plc value: newMethod.
	].
	objectMemory storePointerUnchecked: index + PlcLkupClassIndex ofObject: plc withValue: lkupClass.
	objectMemory storePointerUnchecked: index + PlcNewMethodIndex ofObject: plc withValue: newMethod.
	objectMemory storePointerUnchecked: index + PlcPrimitiveIndex ofObject: plc withValue: (objectMemory integerObjectOf: primitiveIndex).
	objectMemory storePointerUnchecked: index + PlcCacheEntryCounterOffset ofObject: plc withValue: ConstZero.
	self debugStoredInto: plc index: index.! !

!VeryLongJumpInterpreter class methodsFor: 'initialization' stamp: 'ag 1/20/2014 19:39'!
initializeBytecodeTable
	"Overwrite original to add a new bytecode - veryLongJump"

	"Interpreter initializeBytecodeTable"
	"Note: This table will be used to generate a C switch statement."

	BytecodeTable := Array new: 256.
	self table: BytecodeTable from:
	#(
		(  0  15 pushReceiverVariableBytecode)
		( 16  31 pushTemporaryVariableBytecode)
		( 32  63 pushLiteralConstantBytecode)
		( 64  95 pushLiteralVariableBytecode)
		( 96 103 storeAndPopReceiverVariableBytecode)
		(104 111 storeAndPopTemporaryVariableBytecode)
		(112 pushReceiverBytecode)
		(113 pushConstantTrueBytecode)
		(114 pushConstantFalseBytecode)
		(115 pushConstantNilBytecode)
		(116 pushConstantMinusOneBytecode)
		(117 pushConstantZeroBytecode)
		(118 pushConstantOneBytecode)
		(119 pushConstantTwoBytecode)
		(120 returnReceiver)
		(121 returnTrue)
		(122 returnFalse)
		(123 returnNil)
		(124 returnTopFromMethod)
		(125 returnTopFromBlock)
		
		(126 veryLongJump)
		(127 unknownBytecode)

		(128 extendedPushBytecode)
		(129 extendedStoreBytecode)
		(130 extendedStoreAndPopBytecode)
		(131 singleExtendedSendBytecode)
		(132 doubleExtendedDoAnythingBytecode)
		(133 singleExtendedSuperBytecode)
		(134 secondExtendedSendBytecode)
		(135 popStackBytecode)
		(136 duplicateTopBytecode)

		(137 pushActiveContextBytecode)
		(138 pushNewArrayBytecode)
		(139 unknownBytecode)
		(140 pushRemoteTempLongBytecode)
		(141 storeRemoteTempLongBytecode)
		(142 storeAndPopRemoteTempLongBytecode)
		(143 pushClosureCopyCopiedValuesBytecode)

		(144 151 shortUnconditionalJump)
		(152 159 shortConditionalJump)
		(160 167 longUnconditionalJump)
		(168 171 longJumpIfTrue)
		(172 175 longJumpIfFalse)

		"176-191 were sendArithmeticSelectorBytecode"
		(176 bytecodePrimAdd)
		(177 bytecodePrimSubtract)
		(178 bytecodePrimLessThan)
		(179 bytecodePrimGreaterThan)
		(180 bytecodePrimLessOrEqual)
		(181 bytecodePrimGreaterOrEqual)
		(182 bytecodePrimEqual)
		(183 bytecodePrimNotEqual)
		(184 bytecodePrimMultiply)
		(185 bytecodePrimDivide)
		(186 bytecodePrimMod)
		(187 bytecodePrimMakePoint)
		(188 bytecodePrimBitShift)
		(189 bytecodePrimDiv)
		(190 bytecodePrimBitAnd)
		(191 bytecodePrimBitOr)	

		"192-207 were sendCommonSelectorBytecode"
		(192 bytecodePrimAt)
		(193 bytecodePrimAtPut)
		(194 bytecodePrimSize)
		(195 bytecodePrimNext)
		(196 bytecodePrimNextPut)
		(197 bytecodePrimAtEnd)
		(198 bytecodePrimEquivalent)
		(199 bytecodePrimClass)
		(200 bytecodePrimBlockCopy)
		(201 bytecodePrimValue)
		(202 bytecodePrimValueWithArg)
		(203 bytecodePrimDo)
		(204 bytecodePrimNew)
		(205 bytecodePrimNewWithArg)
		(206 bytecodePrimPointX)
		(207 bytecodePrimPointY)

		(208 255 sendLiteralSelectorBytecode)
	).! !

!VeryLongJumpInterpreter methodsFor: 'bytecodes' stamp: 'ag 1/20/2014 19:39'!
veryLongJump
	
	| offset |
	offset := (self fetchByte - 128) * 256 + self fetchByte.
	localIP := localIP + offset.
	offset < 0 ifTrue: [
		"backward jump means we're in a loop; check for possible interrupts"
		self internalQuickCheckForInterrupts.
	].
	self fetchNextBytecode! !
PLCInterpreter initialize!
